"""
This module may still be doing too many different things. I'll probably break it up further in the future.

1. It supplies one means to bind rule-actions to driver-objects. (It would be nice to support a few more styles.)
2. It provides a (maybe) convenient runtime interface to the most common use cases.
"""

import functools, sys
from typing import Iterable

from ..support import failureprone
from . import expansion
from ..parsing import shift_reduce
from ..parsing.interface import HandleFindingAutomaton, END_OF_TOKENS, ParseErrorListener
from ..scanning.interface import FiniteAutomaton, INITIAL, Bindings, RuleId
from ..scanning.engine import IterableScanner
from .interface import ScanAction

class MissingMethodError(TypeError):
	pass

class MacroScanBindings(Bindings):
	def __init__(self, each_action:Iterable[ScanAction], driver):
		def bind(act:ScanAction):
			message_name, *args = act.message
			method_name = 'scan_' + message_name
			try: fn = getattr(driver, method_name)
			except AttributeError:
				if default: fn = functools.partial(default, message_name)
				else: raise MissingMethodError(method_name, act.line_number)
			return act.right_context, fn, args
		default = getattr(driver, 'default_scan_action', None)
		self.__rule = [bind(act) for act in each_action]
		if hasattr(driver, "on_stuck"):
			self.__on_stuck = driver.on_stuck
		else:
			self.__on_stuck = super().on_stuck
	
	def on_match(self, yy, rule_id: RuleId):
		right_context, fn, args = self.__rule[rule_id]
		yy.less(right_context)
		fn(yy, *args)
	
	def on_stuck(self, yy):
		return self.__on_stuck(yy)


def parse_action_bindings(driver, each_constructor):
	"""
	Build a reduction function (combiner) for the parse engine out of an arbitrary Python object.
	Because this checks the message catalog it can be sure the method lookup will never fail.
	"""
	def bind(constructor, mentions):
		if constructor is None: return None
		is_mid_rule_action = constructor.startswith(':')
		kind, constructor = ('mid_rule', constructor[1:]) if is_mid_rule_action else ('parse', constructor)
		specific = kind + '_' + constructor
		try: return getattr(driver, specific)
		except AttributeError:
			generic = 'default_'+kind
			try: method = getattr(driver, generic)
			except AttributeError: raise MissingMethodError(specific, min(mentions))
			else: return functools.partial(method, constructor)
	dispatch = [bind(constructor, mentions) for constructor, mentions in each_constructor]
	def combine(cid:int, args):
		message = dispatch[cid]
		if message is None: return tuple(args) # The null check is like one bytecode and very fast.
		return message(*args)
	return combine


class AbstractTypical(ParseErrorListener):
	"""
	Many applications have similar requirements for error reporting and recovery.
	I'd like to be able to provide a common basis of reasonable default behavior
	for applications that involve both scanning and parsing, regardless of whether
	those applications are founded on the 'mini' or the 'macro' framework.
	
	Specialized versions of this class support either of the two frameworks.
	
	THIS MAY SEEM like a method-as-object (anti)pattern. However, the real
	point is a whole mess of configuration and cooperation in one place.
	"""
	
	source: failureprone.SourceText
	yy: IterableScanner
	exception: Exception
	
	def __init__(self, *, dfa: FiniteAutomaton, bindings, hfa:HandleFindingAutomaton, combine):
		self.__dfa = dfa
		self.__bindings = bindings
		self.__hfa = hfa
		self.__combine = combine
	
	def parse(self, text: str, *, line_breaks='normal', filename: str = None, start=None, language=None):
		if start is None: start = INITIAL
		self.source = failureprone.SourceText(text, line_breaks=line_breaks, filename=filename)
		self.yy = IterableScanner(text, self.__dfa, self.__bindings, start=start)
		return shift_reduce.parse(self.__hfa, self.__combine, self.yy, language=language, on_error=self)
	
	def unexpected_token(self, kind, semantic, pds):
		self.source.complain(*self.yy.slice(), message="Unexpected token %r" % kind)
		# stack_symbols = list(map(self.table.get_breadcrumb, pds.path_from_root()))[1:]
		# self.exception = ParseError(stack_symbols, kind, semantic)
		# print("Parsing condition was:\n", self.exception.condition(), file=sys.stderr)
	
	def unexpected_eof(self, pds):
		self.unexpected_token(END_OF_TOKENS, None, pds)
	
	def will_recover(self, tokens):
		if len(tokens) >= 3: print("Trying to recover.", file=sys.stderr)
		
	def did_not_recover(self):
		print("Could not recover.", file=sys.stderr)
	
	def exception_parsing(self, ex: Exception, message, args):
		self.source.complain(*self.yy.slice(), message="During " + repr(message))
		raise ex from None
	
	def unexpected_character(self, yy: IterableScanner):
		self.source.complain(yy.current_position(), message="Lexical scan got stuck.")
	
	def exception_scanning(self, yy:IterableScanner, rule_id:int, ex:Exception):
		self.source.complain(*yy.slice())
		raise ex from None

def _upgrade_table_0_0_1(tables):
	action = tables['scanner']['action']
	action['argument'] = action.pop('parameter')
	tables['version'] = (0,0,2)

def _upgrade_table_0_0_2(tables):
	action = tables['scanner']['action']
	action['right_context'] = action.pop("trail")
	message = [[m] for m in action['message']]
	for m,a in zip(message, action.pop['argument']):
		if a is not None:
			m.append(a)
	action['message'] = message
	tables['version'] = (0,0,3)


class TypicalApplication(AbstractTypical):
	"""
	This class specializes for the case of compiled tables such as from MacroParse,
	to provide reasonable default error handling behavior.
	"""
	
	def __init__(self, tables, strict=True):
		def version(): return tuple(tables.get('version', ()))
		if version() == (0, 0, 1): _upgrade_table_0_0_1(tables)
		if version() == (0, 0, 2): _upgrade_table_0_0_2(tables)
		if version() != (0, 0, 3):
			raise ValueError('Installed package cannot understand table version: ' + repr(version()))
		
		scanner_table = tables['scanner']
		each_action = expansion.scan_actions(scanner_table['action'])
		dfa = expansion.CompactDFA(dfa=scanner_table['dfa'], alphabet=scanner_table['alphabet'])
		bindings = MacroScanBindings(each_action, driver=self)
		
		parser_table = tables['parser']
		hfa = expansion.CompactHFA(parser_table)
		combine = parse_action_bindings(self, hfa.each_constructor())
		
		super().__init__(dfa=dfa, bindings=bindings, hfa=hfa, combine=combine)





from typing import Optional, NamedTuple, Sequence, Union

class ScanAction(NamedTuple):
	""" The information necessary to connect to a driver (presumably) or yield a usable error message. """
	right_context: Optional[int]
	message: Sequence
	line_number: int




""" Small is beautiful. These algorithms need no introduction. """

from collections.abc import Sequence
from collections import deque
import operator

def allocate(a_list:list, item):
	"""
	Append an item to a list, and return the new item's index in that list.
	Too frequent an idiom not to abbreviate.
	"""
	idx = len(a_list)
	a_list.append(item)
	return idx

def transitive_closure(roots, successors) -> set:
	"""
	Transitive closure is a simple application of graph search.
	(This particular implementation is breadth-first.)
	
	This function does not expect any particular data structure.
	Rather, it takes the graph's outbound-edge relation as a callable parameter.
	It requires:
		``roots`` is an iterable of nodes;
		each node is hashable;
		and ``successors(aNode)`` returns an iterable of nodes.
	"""
	closure = set(roots)
	queue = deque(closure)
	while queue:
		more = successors(queue.popleft())
		if more is not None:
			for item in more:
				if item not in closure:
					closure.add(item)
					queue.append(item)
	return closure

class BreadthFirstTraversal:
	"""
	This object supports more general breadth-first graph traversal (and discovery) algorithms.
	It also accumulates and exposes data about the traversal path, which is usually useful later.
	
	Initialize the traversal's roots by calling .lookup(rootKey) as many times as necessary,
	then perform the traversal by calling .execute(visit). The "visit" parameter must be callable:
	it will be called once with each key this object encounters in a .lookup(...) call.
	In the end, the fields will have these meanings:
	
	current: the index of whichever key is currently being visited; ``None`` before and after processing.
	traversal: the list of keys in the order seen by .lookup(...)
	catalog: the mapping from key to traversal-index
	earliest_predecessor: reverse links pointing along a shortest/first-encountered path towards the root.
	breadcrumbs: Assuming edge-labels are provided with .lookup(key, breadcrumb=label), these are those labels.
	
	"""
	def __init__(self):
		self.current, self.traversal, self.catalog, self.earliest_predecessor, self.breadcrumbs = None, [], {}, [], []
	def execute(self, visit):
		""" visit(key) should call .lookup(successor_key, breadcrumb), which returns an integer. """
		for self.current, key in enumerate(self.traversal):
			visit(key)
		self.current = None
	def lookup(self, key, *, breadcrumb=None) -> int:
		if key not in self.catalog:
			self.catalog[key] = allocate(self.traversal, key)
			self.earliest_predecessor.append(self.current)
			self.breadcrumbs.append(breadcrumb)
		return self.catalog[key]
	def depth_list(self) -> list:
		result = []
		for p in self.earliest_predecessor: result.append(0 if p is None else 1+result[p])
		return result
	def shortest_path_to(self, index:int) -> list:
		""" Return a minimal list of states traversed, from a root to the given node index, in normal order. """
		path = []
		while index is not None:
			path.append(index)
			index = self.earliest_predecessor[index]
		path.reverse()
		return path

class EquivalenceClassifier:
	"""
	Conceivably you might want to normalize the keys,
	but you can do that in advance of the .classify call,
	so there is no point folding it into this object.
	"""
	def __init__(self):
		self.catalog = {}
		self.exemplars = []
	def classify(self, key):
		if key not in self.catalog:
			self.catalog[key] = allocate(self.exemplars, key)
		return self.catalog[key]

def hamming_distance(a, b):
	""" Compute the number of places in two (equal-length) sequences with unequal corresponding elements. """
	assert len(a) == len(b)
	return sum(map(operator.ne, a, b))

def grade(seq:Sequence, *, descending=False) -> list:
	"""
	Returns an index permutation, such that:
		[seq[i] for i in grade(seq)] == sorted(seq)
	Cribbed from APL's grade_up/grade_down primitives.
	"""
	return sorted(range(len(seq)), key=seq.__getitem__, reverse=descending)

def everted(permutation:Sequence) -> list:
	""" Creates and returns the inverse of a given permutation: like it's been turned inside-out """
	result = [None] * len(permutation)
	for i, x in enumerate(permutation): result[x] = i
	return result

def collation(seq:Sequence, *, descending=False):
	""" The inverse of `grade`: Tells you how a sorted sequence was permuted to get here. """
	return everted(grade(seq, descending=descending))

def strongly_connected_components_by_tarjan(graph):
	"""
	See https://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
	Returns a list of strongly-connected components in reverse topological order.
	Each component is a list of member node numbers. Deviating slightly from the wikipedia
	presentation, the low-link is local to the recursive call, eliminating one confusion.
	
	It's expected that graph[q] is the list of arcs from (or perhaps to) node q.
	The linear-time bound assumes all nodes are numbered from 0..last. An isomorphism
	for hashable node keys is provided below.
	"""
	def unvisited(q): return index[q] is None
	def connect(q) -> int:
		low_link = index[q] = allocate(stack, q)
		on_stack[q] = True
		for r in graph[q]:
			if unvisited(r): low_link = min(low_link, connect(r))
			elif on_stack[r]: low_link = min(low_link, index[r])
		if low_link == index[q]:  # i.e. if node q is the root of an SCC:
			component = stack[low_link:]
			del stack[low_link:]
			for r in component: on_stack[r] = False
			output.append(component)
		return low_link
	def main():
		for q in range(size):
			if unvisited(q): connect(q)
	size = len(graph)
	index = [None] * size
	on_stack = [False] * size
	stack = []
	output = []
	main()
	return output

def strongly_connected_components_hashable(graph:dict):
	"""
	Adaptation of Tarjan's SCC algorithm for more kinds of graph node labels than strictly integers.
	The input graph is represented as a dictionary, with values being lists of keys.
	The result will again be lists of keys.
	"""
	table = list(graph.keys())
	index = {key:i for i,key in enumerate(table)}
	prime = [
		[index[arc] for arc in node if arc in index]
		for node in graph.values()
	]
	return [[table[q] for q in component] for component in strongly_connected_components_by_tarjan(prime)]


class Visitor:
	"""
	Visitor-pattern in Python, with fall-back to superclasses along the MRO.

	Actual visitation-algorithms will inherit from Visitor and then each
	`visit_Foo` method must call `self.visit(host.bar)` as appropriate. This
	is so that your visitation-algorithm is in control of which bits of an
	object-graph that it actually visits, and in what order.
	
	Does this really belong here? Perhaps it lacks the gravitas of classical
	algorithms, but it's certainly generic enough and useful in structure-
	directed applications.
	"""
	
	def visit(self, host, *args, **kwargs):
		method_name = 'visit_' + host.__class__.__name__
		try: method = getattr(self, method_name)
		except AttributeError:
			# Searching the MRO incurs whatever cost there is to set up an iterator.
			# NB: Multiple-inheritance with NamedTuple seems to confuse the __mro__.
			for cls in host.__class__.__mro__:
				fallback = 'visit_' + cls.__name__
				if hasattr(self, fallback):
					method = getattr(self, fallback)
					break
			else: raise
		return method(host, *args, **kwargs)

def generate_primes():
	"""
	Generate and yield an unbounded ordered sequence of prime numbers.
	* If you want the first N primes, compose with itertools.islice(..., N).
	* If you want primes less than N, compose with itertools.dropWhile(N.__gt__, ...)

	Various algorithms rely on having a convenient source of prime numbers.
	Although generating primes isn't all that expensive an activity,
	compute-by-need means thinking about thread-safety, so design with care.
	
	There are faster algorithms for large primes, but this wheel-sieve
	(see https://en.wikipedia.org/wiki/Wheel_factorization) is more than
	adequate for general use outside of cryptography and heavy number-theory.
	"""
	yield from (2, 3, 5)  # The basis of the wheel
	
	# Generate a wheel via simple trial division of odd numbers (themselves a sort of minimal wheel):
	# The wheel consists of the non-basis primes less than `2+circumference`.
	# Hard-coded trial division of odd numbers is sufficient to find it:
	offset = circumference = 2 * 3 * 5  # The product of the basis-primes.
	wheel = [n for n in range(7, 2 + circumference, 2) if n % 3 and n % 5]
	yield from wheel
	
	# Henceforth, candidate primes are of the form:
	#     wheel[i] + k * circumference
	# where k ranges from 1..infinity.
	primes, trials, next_trial, next_square = list(wheel), [], 7, 49
	while True:
		for spoke in wheel:
			candidate = spoke + offset
			if candidate >= next_square:
				trials.append(next_trial)
				next_trial = primes[len(trials)]
				next_square = next_trial * next_trial
			if all(candidate % p for p in trials):
				yield candidate
				primes.append(candidate)
		offset += circumference




"""
Of particular importance to my cause is a proper bootstrap.
For example, I could hand-roll a recursive-descent parser for regular expressions,
but that would be at odds with the presumed benefits of parser-generation machinery.
The approach taken here is to extract a full-strength system nearly for free.
"""

from typing import NamedTuple, Optional, Any
from ..support import foundation
from ..support.failureprone import SourceText, Issue, Evidence, Severity, illustration
from ..support.treelang import RankedAlphabet, StrictPass, SparseRewrite
from ..support.symtab import NameSpace, NoSuchSymbol
from .interface import Bindings, RuleId
from . import charset
from .finite import NFA
from .engine import IterableScanner
from ..parsing.context_free import ContextFreeGrammar, Rule, SemanticAction
from ..parsing.interface import ParseError
from ..parsing import shift_reduce
from ..parsing.lalr import lalr_construction
from ..parsing.automata import tabulate, DeterministicStyle

class PatternError(Exception): pass
class VariableTrailingContextError(PatternError): pass
class NameTooShort(Exception): pass


VOCAB = RankedAlphabet("Rule", "Char", "Cls", "Regex", "LeftContext", "Number", "Codepoint")

VOCAB.symbol("pattern_regular", "Rule", left_context="LeftContext", stem="Regex")
VOCAB.symbol("pattern_with_right_context", "Rule", left_context="LeftContext", stem="Regex", right_context="Regex")
VOCAB.symbol("pattern_only_right_context", "Rule", left_context="LeftContext", right_context="Regex")

VOCAB.symbol("sequence", "Regex", a="Regex", b="Regex")
VOCAB.symbol("alternation", "Regex", a="Regex", b="Regex")

VOCAB.symbol("star", "Regex", sub="Regex")
VOCAB.symbol("hook", "Regex", sub="Regex")
VOCAB.symbol("plus", "Regex", sub="Regex")

VOCAB.symbol("n_times", "Regex", sub="Regex", count="Number")
VOCAB.symbol("n_or_more", "Regex", sub="Regex", min="Number")
VOCAB.symbol("n_or_fewer", "Regex", sub="Regex", max="Number")
VOCAB.symbol("n_to_m", "Regex", sub="Regex", min="Number", max="Number")

VOCAB.symbol("cls", "Regex", members="Cls")
VOCAB.symbol("initial_minus", "Cls")

VOCAB.symbol("singleton", "Cls", item="Codepoint")
VOCAB.symbol("range", "Cls", first="Codepoint", last="Codepoint")
VOCAB.symbol("union", "Cls", a="Cls", b="Cls")
VOCAB.symbol("intersection", "Cls", a="Cls", b="Cls")
VOCAB.symbol("subtraction", "Cls", a="Cls", b="Cls")
VOCAB.symbol("complement", "Cls", inverse="Cls")

# Terminals from here down:
VOCAB.symbol("epsilon", "Regex")
VOCAB.symbol("name", "Regex", "Cls")
VOCAB.symbol("shorthand", "Regex", "Cls")
VOCAB.symbol("literal", "Regex", "Codepoint")
VOCAB.symbol("end", "Regex", "Codepoint")
VOCAB.symbol("escape", "Regex", "Codepoint")
VOCAB.symbol("control", "Regex", "Codepoint")
VOCAB.symbol("hex_point", "Regex", "Codepoint")
VOCAB.symbol("number", "Number")
VOCAB.symbol("dot", "Regex")

VOCAB.symbol("anywhere", "LeftContext")
VOCAB.symbol("begin_line", "LeftContext")
VOCAB.symbol("mid_line", "LeftContext")

# Internal-Use-Only Symbols:
VOCAB.symbol("codepoint_", "Regex", "Codepoint", trace="Codepoint")

#################################################
#
#  Now I can describe a translation strategy.
#
# 1. Convert all literals, escapes, controls, and hex-points to their respective codepoint ordinals.
#    This may imply adding a generic symbol "codepoint" and subtracting other members of the category.
# 2. Evaluate numbers. Check that n_to_m items make sense, in the sense of min <= max.
# 3. Look up names and shorthands from the environment and check for acceptability within character classes.
#    - Everything from here up involves taking extracts from the text. Henceforth, we don't need it.
# 4. Adjust for right-context in rule-pattern, which may involve checking the sizes of stem and right_context.
# --- At this point, you can drop positional information: Conversion can no longer fail. ---
# 5. Do something about the alphabet. For example, turn singletons into trivial character classes.
# 6. Optionally, bottom-up, convert alternations-of-classes into single classes. This may save time later.
# 7. The actual NFA encoding process should follow the advice at https://swtch.com/~rsc/regexp/regexp1.htm
#    To wit: Treat the destination-state as an inherited attribute, and the source-state as a synthesized one.
# 8. The usual procedure to make a minimal DFA with a minimal alphabet.
#
# There is some freedom to shift and merge portions of this stuff. For example, to add case-folding
# as a flag on the parentheses, I'd use an inherited attribute to control converting literal letters.

class Annotation(NamedTuple):
	codepoints: dict[object:int]
	names: dict[object:]

class RulePattern(NamedTuple):
	bol: tuple[bool, bool]
	tree: Any
	trail_code: Optional[int]
	annotation: Annotation

class DefinitionEntry(NamedTuple):
	tree: Any
	size: Optional[int]
	annotation: Annotation

class ConvertLiterals(SparseRewrite):
	def __init__(self, *, text:str, slices:dict, env:NameSpace):
		self.__text = text
		self.__slices = slices
		self.__env = env
		assert isinstance(env, NameSpace)
		
		# Public attributes will be the results of analysis:
		self.codepoints = {}
		self.names = {}
		self.numbers = {}
		self.warnings = {}
		self.errors = {}
	
	def _mk(self, node, ordinal):
		literal = VOCAB['codepoint_'](node)
		self.codepoints[literal] = ordinal
		return literal
	
	def literal(self, node, in_class=False):
		return self._mk(node, ord(self.__text[self.__slices[node]]))

	def escape(self, node, in_class=False):
		return self._mk(node, ord(self.__text[self.__slices[node].stop-1]))

	def control(self, node, in_class=False):
		return self._mk(node, ord(self.__text[self.__slices[node].stop-1]) % 32)
	
	def hex_point(self, node, in_class=False):
		the_slice = self.__slices[node]
		the_text = self.__text[the_slice.start+2:the_slice.stop]
		return self._mk(node, int(the_text, 16))

	def number(self, node):
		the_text = self.__text[self.__slices[node]]
		self.numbers[node] = int(the_text)
		return node
	
	def n_to_m(self, node):
		sub = self(node.sub)
		least = self(node.min)
		most = self(node.max)
		if self.numbers[least] > self.numbers[most]:
			self.warnings[node] = "Backwards Boundaries; swapping..."
			S = self.__slices
			S[node] = slice(S[node.min].start, S[node.max].stop)
			least, most = most, least
		node = VOCAB['n_to_m'](sub, most, least)
		return node
	
	def cls(self, node):
		""" Inherited attribute useful to make sure defined char-classes get used the right way. """
		return self._unhandled_(node, True)
	
	def __lookup(self, node, in_class, text):
		try: entry = self.names[node] = self.__env[text]
		except NoSuchSymbol: self.errors[node] = "No such symbol as %r in this environment."%text
		else:
			if in_class and type(entry) is not list and type(entry.tree) is not VOCAB['cls']:
				self.errors[node] = "Defined name %r does not refer to a character class here."%text
			return node
	
	def name(self, node, in_class=False):
		s = self.__slices[node]
		return self.__lookup(node, in_class, self.__text[s.start + 1:s.stop - 1])
	
	def shorthand(self, node, in_class=False):
		proxy = VOCAB['name']()
		text = self.__text[self.__slices[node].start+1]
		return self.__lookup(proxy, in_class, text)
	
	def end(self, node):
		# Convert to being a name-reference. Can't happen inside a char-class.
		proxy = VOCAB['name']()
		self.names[proxy] = charset.EOL
		return proxy

	def dot(self, node):
		# Convert to being a name-reference. Can't happen inside a char-class.
		proxy = VOCAB['name']()
		self.names[proxy] = charset.DOT
		return proxy

class RuleAnalyzer(StrictPass):
	""" Just enough to extract a pattern's left and right context information. """
	
	def __init__(self, names):
		# By now I'm assuming names have been resolved,
		# so I don't need the environment.
		self.__names = names

	def pattern_regular(self, node):
		# No trailing context, so trail is None
		return self(node.left_context), node.stem, None
	
	def pattern_with_right_context(self, node):
		stemSize, trailSize = self(node.stem), self(node.right_context)
		expression = VOCAB['sequence'](node.stem, node.right_context)
		if trailSize:
			trailCode = - trailSize
		elif stemSize:
			trailCode = stemSize
		else:
			raise VariableTrailingContextError(
				'Variable stem and variable trailing context in the same pattern are not presently supported.')
		return self(node.left_context), expression, trailCode
	
	def pattern_only_right_context(self, node):
		return self(node.left_context), node.right_context, 0
	
	def anywhere(self, n):
		return (True, True)
	
	def begin_line(self, n):
		return (False, True)
	
	def mid_line(self, n):
		return (True, False)
	
	def cls(self, node):
		return 1
	
	def codepoint_(self, node):
		return 1
	
	def sequence(self, node):
		a = self(node.a)
		if a is not None:
			b = self(node.b)
			if b is not None:
				return a+b
			
	def alternation(self, node):
		a = self(node.a)
		if a is not None:
			b = self(node.b)
			if b == a:
				return b

	def star(self, _): return None
	def hook(self, _): return None
	def plus(self, _): return None
	def epsilon(self, _): return 0
	
	def name(self, node):
		entry = self.__names[node]
		return 1 if type(entry) is list else entry.size

class RemoveCounts(SparseRewrite):
	"""
	Counted notation is convenient shortcut.
	To implement it, I'll also take a shortcut:
	Specific forms of count-expressions get rewritten in terms of sequence, star, hook, and epsilon.
	There's an interesting consequence:
	The resulting "tree" is no longer a tree, but a DAG.
	The algorithms won't care.
	"""
	def __init__(self, numbers):
		self.__numbers = numbers
	def n_times(self, n):
		count = self.__numbers[n.count]
		return self.__convert(n.sub, count, count)
	def n_to_m(self, n):
		least = self.__numbers[n.min]
		most = self.__numbers[n.max]
		return self.__convert(n.sub, least, most)
	def n_or_more(self, n):
		least = self.__numbers[n.min]
		return self.__convert(n.sub, least, None)
	def n_or_fewer(self, n):
		most = self.__numbers[n.max]
		return self.__convert(n.sub, 0, most)
	
	@staticmethod
	def __convert(sub, least:int, most):
		"""
		This will return a sequence (or an epsilon node, as the case may be).
		The first ``least`` elements will be references to sub.
		Then, ``most - least`` elements will be optional references.
		Or, if ``most`` is unbounded, the last element us made a plus-reference.
		However, I wrote it in reverse.
		"""
		if most is None:
			it = VOCAB['star'](sub)
		else:
			it = VOCAB['epsilon']()
			option = VOCAB['hook'](sub)
			for _ in range(least, most):
				it = VOCAB['sequence'](option, it)
		for _ in range(least):
			it = VOCAB['sequence'](sub, it)
		return it

class Encoder(StrictPass):
	"""
	This represents the strategy to encode a regular expression as
	a non-deterministic finite state automaton.
	Cribbing from https://swtch.com/~rsc/regexp/regexp1.htm
	"""
	def __init__(self, nfa:NFA, *, annotation:Annotation, rank:int):
		self.__nfa = nfa
		self.__rank = rank
		self.__ann = annotation
		self.__ce = ClassEncoder(names=annotation.names, codepoints=annotation.codepoints)
	def __new_node(self) -> int: return self.__nfa.new_node(self.__rank)
	def __eps(self, src:int, dst:int, ): self.__nfa.link_epsilon(src, dst)
	def codepoint_(self, n, dst:int, ) -> int:
		src = self.__new_node()
		self.__nfa.link(src, dst, charset.singleton(self.__ann.codepoints[n]))
		return src
	
	def cls(self, n, dst:int) -> int:
		compiled_class = self.__ce(n.members)
		src = self.__new_node()
		self.__nfa.link(src, dst, compiled_class)
		return src
	
	@staticmethod
	def epsilon(_, dst:int, ) -> int:
		return dst
	
	def alternation(self, alt, dst:int, ) -> int:
		src = self.__new_node()
		self.__eps(src, self(alt.a, dst))
		self.__eps(src, self(alt.b, dst))
		return src
	
	def sequence(self, seq, dst:int, ) -> int:
		return self(seq.a, self(seq.b, dst))
	
	def hook(self, h, dst:int, ) -> int:
		src = self.__new_node()
		self.__eps(src, self(h.sub, dst))
		self.__eps(src, dst)
		return src
		
	def star(self, s, dst:int, ) -> int:
		src = self.__new_node()
		self.__eps(src, self(s.sub, src))
		self.__eps(src, dst)
		return src
		
	def plus(self, p, dst:int, ) -> int:
		mid = self.__new_node()
		src = self(p.sub, mid)
		self.__eps(mid, src)
		self.__eps(mid, dst)
		return src
		

	def __name(self, ns, dst:int, ) -> int:
		dfn = self.__ann.names[ns]
		if type(dfn) is list:
			src = self.__new_node()
			self.__nfa.link(src, dst, dfn)
			return src
		else:
			enc = Encoder(self.__nfa, rank=self.__rank, annotation=dfn.annotation)
			return enc(dfn.tree, dst)

	name = shorthand = __name


class ClassEncoder(StrictPass):
	"""
	Builds a character class in the format used by the finite.NFA class.
	
	One might think this could all be done inside class Encoder, but the
	treatment of name references differs.
	"""
	def __init__(self, *, names:dict, codepoints:dict):
		self.__names = names
		self.__codepoints = codepoints
	
	def initial_minus(self, node):
		return charset.singleton(ord("-"))
	def __name(self, n):
		# By this point, the name is proven safe to use in a class.
		# That check happens before a definition goes into the symbol table.
		dfn : DefinitionEntry = self.__names[n]
		if type(dfn) is list:
			return dfn
		else:
			ann = dfn.annotation
			ce = ClassEncoder(names=ann.names, codepoints=ann.codepoints)
			return ce(dfn.tree.members)
	name = shorthand = __name
	def singleton(self, n): return charset.singleton(self.__codepoints[n.item])
	def range(self, n): return charset.range_class(self.__codepoints[n.first], self.__codepoints[n.last])
	def union(self, n): return charset.union(self(n.a), self(n.b))
	def intersection(self, n): return charset.intersect(self(n.a), self(n.b))
	def complement(self, n): return charset.complement(self(n.inverse))
	def subtraction(self, n): return charset.intersect(self(n.a), charset.complement(self(n.b)))
	def empty_set(self, n): return charset.EMPTY


def annotate(text, env, language) -> tuple[Any, [Annotation]]:
	slices = {}
	tree_1 = parse_regex(text, slices, language=language)
	pass_1 = ConvertLiterals(text=text, slices=slices, env=env)
	tree_2 = pass_1(tree_1)
	
	if pass_1.warnings or pass_1.errors:
		st = SourceText(text)
		for node, error in pass_1.warnings.items():
			Issue("Pass-1", Severity.WARNING, error, {None: [Evidence(slices[node])]}).emit(lambda _: st)
		for node, error in pass_1.errors.items():
			Issue("Pass-1", Severity.ERROR, error, {None: [Evidence(slices[node])]}).emit(lambda _: st)
		if pass_1.errors:
			raise PatternError(text)
	
	tree_3 = RemoveCounts(pass_1.numbers)(tree_2)
	return tree_3, Annotation(pass_1.codepoints, pass_1.names)


def analyze_pattern(pattern_text, env) -> Optional[RulePattern]:
	tree, ann = annotate(pattern_text, env, 'Pattern')
	pass_2 = RuleAnalyzer(ann.names)
	bol, expression, trail_code = pass_2(tree)
	return RulePattern(bol, expression, trail_code, ann)


def let_subexpression(env, name, pattern):
	assert isinstance(name, str)
	if len(name) <= 1: raise NameTooShort
	tree, ann = annotate(pattern, env, 'Regular')
	size = RuleAnalyzer(ann.names)(tree)
	env[name] = DefinitionEntry(tree, size, ann)

############

REGEX_SUGAR = set(r"^ / \ | ? * + { } , ( ) [ [^ - ] && &&^ ; whitespace".split())

def install_regex_grammar(grammar:ContextFreeGrammar, message_mapper=lambda x:x):
	"""
	Problem one: Context-free grammar for regular expressions.
	The object is to add rules to a CFG representing how to
	parse a regular expression from its constituent tokens.
	
	This gets extracted in hopes of reusing it in a smarter meta-compiler later.
	"""
	for line in """
		Pattern: left_context Regular               :pattern_regular
		Pattern: left_context Regular right_context :pattern_with_right_context
		Pattern: left_context right_context         :pattern_only_right_context
		left_context:    :anywhere
		left_context: ^  :begin_line
		left_context: ^ ^ :mid_line
		right_context: end
		right_context: / Regular
		right_context: / Regular end :sequence
		Regular: sequence
		Regular: Regular | sequence :alternation
		sequence: term
		sequence: sequence term :sequence
		term: atom
		term: atom ? :hook
		term: atom * :star
		term: atom + :plus
		term: atom { number }          :n_times
		term: atom { number , }        :n_or_more
		term: atom { , number }        :n_or_fewer
		term: atom { number , number } :n_to_m
		atom: codepoint
		atom: name
		atom: shorthand
		atom: dot
		atom: ( Regular )
		atom: class ] :cls
		class: [ conjunct
		class: [^ conjunct :complement
		class: class && conjunct :intersection
		class: class &&^ conjunct :subtraction
		conjunct: item
		conjunct: - :initial_minus
		conjunct: conjunct item :union
		item: codepoint :singleton
		item: codepoint - codepoint :range
		item: name
		item: shorthand
		codepoint : literal
		codepoint : escape
		codepoint : control
		codepoint : hex_point
	""".splitlines():
		bits = [x.strip() for x in line.split(':')]
		if len(bits) < 2: continue
		lhs, rhs = bits[0], tuple(bits[1].split())
		indices = [i for i, s in enumerate(rhs) if s not in REGEX_SUGAR]
		if len(bits) == 2:
			assert len(indices) == 1, line
			action = indices[0]
		elif len(bits) == 3:
			action = SemanticAction(message_mapper(bits[2]), indices)
		else: assert False
		grammar.add_rule(Rule(lhs, rhs, None, action, line))


def declare_regex_token_rules(nfa, make_rule):
	"""
	A recognizer for regex tokens.

	One approach is to build an NFA directly from state transition rules.
	This implies a simplified (internal) DSL for such rules.
	In this DSL, every character will stand for itself except:
		normal shorthand-escape letters serve their purpose without an escape symbol.
		the underscore shall stand in for the plus symbol (because we need to recognize plus, unescaped).
		C for the char range from @ to ~ in the ASCII space. (i.e. 64 - 126)
	"""
	def connect(cond, pattern, rule_id):
		src = prior = nfa.new_node(0)
		for q in nfa.condition(cond): nfa.link_epsilon(q, src)
		for c in pattern:
			if c == '_':
				nfa.link_epsilon(src, prior)
			else:
				if c in bootclass: label = bootclass[c]
				else: label = charset.singleton(ord(c))
				dst = nfa.new_node(0)
				nfa.link(src, dst, label)
				prior, src = src, dst
		nfa.final[src] = rule_id
	def declare(cond, pattern, kind, then=None):
		rule_id = make_rule(kind, then)
		connect(cond, pattern, rule_id)
	def meta(cond, token, then=None):
		return declare(cond, token, token, then)
	
	bootclass = charset.mode_normal.new_child("Bootstrap Extras")
	bootclass['C'] = charset.range_class(64, 127)
	bootclass['%'] = charset.POSIX['xdigit']
	
	ignore =  make_rule(None, None)
	for cond in 'group', 'brace', 'class':
		connect(cond, 's_', ignore)
	meta("group", ")", "POP")
	meta("init", "(", "group")
	meta("init", "[", "class")
	meta("init", '[^', "class")
	meta("init", '{', "brace")
	declare("init", '.', 'dot')
	for c in "^$/": declare("group", c, 'literal')
	for c in r'|}])/^\?*+': meta("init", c)
	declare("init", '$', 'end')

	declare("brace", 'd_', "number")
	meta("brace", ',')
	meta("brace", '}', "POP")
	
	meta('class', '-')
	meta('class', '&&')
	meta('class', '&&^')
	meta('class', ']', "POP")

	declare("esc", r'\L', "escape")
	declare("esc", '{lw_}', "name")
	declare("esc", r"\cC", 'control')
	declare("esc", r"\l", 'shorthand')
	hex_point = make_rule("hex_point", None)
	for pattern in (r"\x%%",  r"\u%%%%", r"\U%%%%%%%%", ):
		connect("esc", pattern, hex_point)
	declare("esc", 'S', 'literal')
	
	nfa.link_condition("group", "init")
	nfa.link_condition("init", "esc")
	nfa.link_condition("class", "esc")

def _regex_scanner():
	nfa = NFA()
	actions = []
	def rule(msg, *args):
		return foundation.allocate(actions, ("scan_"+msg, args))
	def makeRule(kind:str, then):
		if kind is None: return rule("ignore")
		elif then is None:
			if kind in REGEX_SUGAR: return rule("sugar", kind)
			else: return rule("token", kind)
		elif then == "POP": return rule("popping", kind)
		else: return rule("pushing", kind, then)
	
	declare_regex_token_rules(nfa, makeRule)
	dfa = nfa.subset_construction().minimize_states().minimize_alphabet()
	return dfa, actions
_DFA, _ACTIONS = _regex_scanner()

class RXBindings(Bindings):
	
	def __init__(self, slices):
		self.__slices = slices
	
	def on_match(self, yy, rule_id:RuleId):
		message, args = _ACTIONS[rule_id]
		method = getattr(self, message)
		method(yy, *args)
	
	def scan_ignore(self, yy):
		pass
	def scan_sugar(self, yy, kind):
		yy.token(kind, None)
	def scan_token(self, yy, kind):
		semantic = VOCAB[kind]()
		self.__slices[semantic] = yy.slice()
		yy.token(kind, semantic)
	def scan_pushing(self, yy, kind, then):
		yy.token(kind, None)
		yy.push(then)
	def scan_popping(self, yy, kind):
		yy.token(kind, None)
		yy.pop()

def _regex_parser():
	rex = ContextFreeGrammar()
	rex.start.extend(['Pattern', 'Regular'])
	install_regex_grammar(rex, message_mapper=VOCAB.__getitem__)
	rex.validate()
	table = tabulate(lalr_construction(rex), rex, style=DeterministicStyle(strict=True))
	constructors = table.constructors
	combine = lambda cid,args:constructors[cid](*args)
	return table, combine
_PARSE, _COMBINE = _regex_parser()


def scan_regex(text, slices):
	return IterableScanner(text, _DFA, RXBindings(slices), "init")

def parse_regex(text, slices, *, language="Pattern", on_error=None):
	each_token = scan_regex(text, slices)
	try:
		return shift_reduce.parse(_PARSE, _COMBINE, each_token, language=language, on_error=on_error)
	except ParseError:
		print("Cannot understand regex:")
		print(illustration(text, each_token.left))
		raise



from collections import defaultdict, Counter
from typing import NamedTuple
from ..support.foundation import transitive_closure, BreadthFirstTraversal
from .interface import END_OF_TOKENS
from .context_free import ContextFreeGrammar, RIGHT
from .automata import HFA, LookAheadState, reachable
from .lr0 import lr0_construction, ParseItemMap
from .lalr import find_lalr_sets

def canonical_lr1(grammar: ContextFreeGrammar) -> HFA[LookAheadState]:
	"""
	Before embarking on a quest to produce a minimal-LR(1) table by sophisticated
	methods, it's worth learning how to produce the maximal-LR(1) table by (some
	variant of) Donald E. Knuth's original method.

	A Knuth parse-item is like an LR(0) item augmented with the (1) next token
	expected AFTER the corresponding rule would be recognized. The initial core
	would look like { .S/# } in the usual notation. Otherwise, the algorithm has
	much in common with the LR(0) construction above -- but to see that clearly
	you'll have to look at the function `abstract_lr1_construction(...)`.
	"""
	
	def front(symbol, follower, goto_transparent, iso_q):
		"""
		Return the list of parse-items which DIRECTLY follow from the given criteria.
		In Canonical LR(1), that means the beginnings of each rule for the
		prediction-symbol, with followers being every terminal that might come
		next after the prediction-symbol got shifted.
		"""
		iso_state = lr0.graph[iso_q]
		goto_state = lr0.graph[iso_state.shift[symbol]]
		after = set(goto_state.shift.keys()) & terminals
		if goto_transparent: after.add(follower)
		return [
			(lr0_item, lookahead)
			for lr0_item in pim.symbol_front.get(symbol)
			for lookahead in after
		]
	
	pim = ParseItemMap.from_grammar(grammar)
	lr0 = lr0_construction(pim)  # This implicitly solves a lot of sub-problems.
	terminals = grammar.apparent_terminals()
	return abstract_lr1_construction(
		pim, grammar,
		front=front,
		note_reduce=lambda reduce, follower, rule_id, iso_q: reduce[follower].append(rule_id),
		initial_follow=END_OF_TOKENS,
		lr0_catalog=lr0.bft.catalog,
	)


def abstract_lr1_construction(pim: ParseItemMap, grammar: ContextFreeGrammar, *, front, note_reduce, initial_follow, lr0_catalog) -> HFA[LookAheadState]:
	"""
	The Canonical and Minimal LR(1) algorithms given here have a great deal in common.
	It seems both instructive and useful to factor out those commonalities.
	If you stare intently, you'll also see similarity to LR(0), but factoring
	out that particular commonality is not today's exercise.
	"""
	
	def build_state(lr1_core: frozenset):
		iso_q = lr0_catalog[frozenset(i for (i, f) in lr1_core)]
		shifted_cores, reduce = defaultdict(set), defaultdict(list)
		
		def visit(lr1_item):
			lr0_item, follower = lr1_item
			
			next_symbol = pim.symbol_at[lr0_item]
			if next_symbol is None:
				rule_id = pim.rule_found[lr0_item]
				if rule_id < len(grammar.rules):
					note_reduce(reduce, follower, rule_id, iso_q)
			else:
				shifted_cores[next_symbol].add((lr0_item + 1, follower))
				if next_symbol in grammar.symbol_rule_ids:
					# i.e. next_symbol is a non-terminal:
					return front(next_symbol, follower, transparent[lr0_item + 1], iso_q)
		
		transitive_closure(lr1_core, visit)
		shift = {
			symbol: bft.lookup(frozenset(item_set), breadcrumb=symbol)
			for symbol, item_set in reachable(shifted_cores, reduce, grammar).items()
		}
		graph.append(LookAheadState(shift=shift, reduce=reduce))
	
	transparent = find_transparent(grammar.find_epsilon(), pim.symbol_at)
	bft = BreadthFirstTraversal()
	graph = []
	initial_items = [(i, initial_follow) for i in pim.language_front]
	initial = [bft.lookup(frozenset([item])) for item in initial_items]
	bft.execute(build_state)
	accept = [graph[qi].shift[language] for qi, language in zip(initial, grammar.start)]
	# print("LR(0) states: %d\t\tLR(1) states:%d" % (len(lr0_catalog), len(graph)))
	return HFA(graph=graph, initial=initial, accept=accept, bft=bft)


def find_transparent(epsilon: set[str], symbol_at:list[str]) -> list[bool]:
	"""
	In LR(1)-family algorithms, we do a fair amount of tests for the rest-of-a-rule
	being transparent: i.e. all the remaining symbols in a right-hand-side (if any)
	being possibly-epsilon. The naive way -- something like
			all(symbol in epsilon for symbol in rhs[position:])
	works just fine, but think of the all those garbage list copies!
	We can do better. This is one way.
	"""
	
	flag = False
	transparent = [False] * len(symbol_at)
	for i in reversed(range(len(transparent))):
		if symbol_at[i] is None:
			flag=True
			transparent[i] = flag
		elif symbol_at[i] in epsilon:
			transparent[i] = flag
		else: flag=False
	return transparent


def minimal_lr1(grammar: ContextFreeGrammar) -> HFA[LookAheadState]:
	"""
	This amounts to a hybrid of LALR and Canonical LR(1) in which only the conflicted
	parts are reconsidered in greater detail. Details of the approach are in the
	doc-comments for the `front` and `note_reduce` subroutines.

	After poring over the IELR paper on several occasions, I believe there may yet be
	some originality in this contribution. If anything, I have an argument why this
	routine produces absolutely minimal tables: each output state has an absolutely
	minimal set of "siblings" affiliated with a corresponding LALR state, because the
	only thing distinguishing "siblings" is the correct final deterministic parse
	action after the rule associated with a parse-item is recognized, and then only
	for those (generally very few) tokens for which LALR does not figure it out. Also,
	unit rule elimination is applied, and unreachable states (in light of conflict
	resolutions) are never considered.
	"""
	
	def front(symbol, follower, goto_transparent, iso_q):
		"""
		Return the list of parse-items which DIRECTLY follow from the given criteria.
		In other words, these will be parse-items in position zero for the rules that
		apply to nonterminal :param symbol:.

		:param follower: may be `None` to mean "the non-conflicted portion of the reduce set,
		or may be a specific reduce-set token.

		:param goto_transparent: means there is an epsilon-only path through any
		remaining portion of the rule :param symbol: appeared in, so that reduce-set
		conflicts should be propagated to these resulting parse-items.

		:param iso_q: is the ID number in the LR(0)/LALR graph corresponding to the
		LR(1) state under construction.

		The complexity here comes from how this algorithm threads the needle between
		LALR-when-adequate and LR(1)-when-necessary.
		"""
		isostate = lr0.graph[iso_q]
		items = []
		goto_q = isostate.shift[symbol]
		goto_conflict = conflict_data[goto_q].tokens.keys()
		for sub_rule_id, lr0_item in zip(grammar.symbol_rule_ids[symbol], pim.symbol_front[symbol]):
			# Most of the smarts in this algorithm comes down to understanding what
			# LALR found at the far end of each sub-production. We need to know which
			# LR(0) state you reach after shifting the contents of that sub-rule.
			# (Incidentally, the LALR construction also discovers this same information.)
			reach = lr0.traverse(iso_q, grammar.rules[sub_rule_id].rhs)
			if follower is None:  # We're coming from LALR-land:
				items.append((lr0_item, None))
				reach_conflict = conflict_data[reach].rules.get(sub_rule_id, EMPTY)
				possible_follow = reach_conflict & successors[goto_q]
				# Things get a bit weird for tokens that are ALSO conflicted in the
				# goto state. Normally, we ignore them in this section; they'll come
				# along expressly in another round through the algorithm as a split
				# from the goto-state. However, in case of epsilon productions we
				# must include those tokens lest the parse table may come out wrong.
				if reach != iso_q: possible_follow -= goto_conflict
				for token in possible_follow: items.append((lr0_item, token))
			else:  # The canonical branch:
				# GOTO-conflicted tokens will have resulted in canonical-style parse items.
				# As with Canonical, they can follow a derivation only when the remainder
				# of the current rule is "transparent", but this algorithm imposes the
				# additional constraint regarding the token's contribution to a
				# LALR-inadequacy in the "reach" state.
				if follower in conflict_data[reach].tokens and goto_transparent:
					assert follower in goto_conflict
					items.append((lr0_item, follower))
		return items
	
	def note_reduce(reduce, follower, rule_id, iso_q):
		"""
		There are two cases:

		If the "follower" is `None`, it stands for the un-conflicted portion of the
		corresponding LALR reduce-set.

		Otherwise, the token MUST have earlier been implicated in a LALR-inadequacy
		in this state (which fact we assert for good measure). Handle it the same as
		Canonical-LR(1).

		Incidentally, it is possible to reach a particular `reduce[follower]` list
		more than once if and only if the follower is LALR-inadequate. Proof follows
		from the fact that a given parse-item is visited at most once.
		"""
		if follower is None:
			for t in terminal_sets[reduce_set_id[iso_q, rule_id]] - conflict_data[iso_q].rules[rule_id]:
				assert t not in reduce
				reduce[t] = [rule_id]
		else:
			assert follower in conflict_data[iso_q].rules[rule_id]
			if follower in reduce:
				assert rule_id not in reduce[follower]
				reduce[follower].append(rule_id)
			else:
				reduce[follower] = [rule_id]
	
	def possible_next_terminals(iso_q):
		union = set(terminal_sets[iso_q])
		for rule_id in lr0.graph[iso_q].reduce:
			union.update(terminal_sets[reduce_set_id[iso_q, rule_id]])
		return union
	
	EMPTY = frozenset()
	pim = ParseItemMap.from_grammar(grammar)
	lr0 = lr0_construction(pim)  # This implicitly solves a lot of sub-problems.
	terminal_sets, reduce_set_id = find_lalr_sets(lr0, grammar)
	successors = [possible_next_terminals(iso_q) for iso_q in range(len(lr0.graph))]
	# Later we need to know if a certain rule is implicated in an LALR conflict: if so, for which terminals?
	# We also need to know if a state is conflicted with respect to a particular terminal.
	conflict_data = find_conflicts(lr0.graph, {(q, r): terminal_sets[i] for (q, r), i in reduce_set_id.items()}, grammar)
	
	return abstract_lr1_construction(
		pim, grammar,
		front=front, note_reduce=note_reduce,
		initial_follow=None,
		lr0_catalog=lr0.bft.catalog,
	)



class ConflictData(NamedTuple):
	tokens: dict[str, set[int]]  # The rules that conflict on this token
	rules: dict[int, set[str]]  # The tokens that conflict on this rule.


def find_conflicts(graph, reduce_sets, grammar) -> list[ConflictData]:
	"""
	This drives one of the central ideas of the Minimal-LR(1) algorithm:
	Learn which tokens are involved in conflicts, and which rules contribute
	to those conflicts for each token (as known to LALR).

	Subtleties:

	1. If an S/R conflict is DECLARED to shift, then it does not impugn the token,
	   but the token still refers to the rule in case some other rule MAY reduce.

	This routine (and the stuff that uses it) is coded with the idea that the
	grammar may leave certain things deliberately non-deterministic. Wherever that
	is the case, these algorithms will respect it.

	2. There is a way to improve the treatment of R/R conflicts if it is known in
	   advance that the table will be used deterministically, or if the R/R is
	   resolved by rule precedence. It involves a pass over the LR(1) item cores
	   considering the groups that eventually lead to a R/R conflict (they have the
	   same suffix and follower): among those groups only the "winning" reduction
	   item needs to stay in the core. This "normalizes" the LR(1) cores so that
	   potentially fewer distinct ones might be generated.

	Alas, idea #2 is not yet implemented. Complex R/R conflicts may still lead to
	more states than strictly necessary for a deterministic table. In practice,
	this is unlikely to be a real problem: deterministic tables are usually made
	from grammars with few LALR conflicts, and in the non-deterministic case
	nothing is wasted. Nevertheless, this remains an avenue for improvement.
	"""
	result = []
	for q, state in enumerate(graph):
		degree = Counter(state.shift.keys())  # This picks up some nonterminals but they do no harm.
		for rule_id in state.reduce:
			for token in reduce_sets[q, rule_id]:
				prefer_shift = token in state.shift and grammar.decide_shift_reduce(token, rule_id) == RIGHT
				if not prefer_shift: degree[token] += 1
		conflicted_tokens = set(token for token, count in degree.items() if count > 1)
		conflict = ConflictData({token: set() for token in conflicted_tokens}, {})
		for rule_id in state.reduce:
			contribution = conflicted_tokens & reduce_sets[q, rule_id]
			conflict.rules[rule_id] = contribution
			for token in contribution: conflict.tokens[token].add(rule_id)
		result.append(conflict)
	return result




"""
# Context Free Grammars

In pure form, a context free grammar (CFG) consists of:
	* A set of terminal symbols,
	* A set of non-terminal symbols, disjoint from the terminals,
	* A set of production rules, each consisting of:
		* left-hand side (exactly one symbol)
		* right-hand side (ordered sequence of zero or more symbols)
	* and a start symbol.

As a cool and potentially useful hack, I expand the normal definition to allow
potentially several "start symbols" to share the same overall definition. The
practical benefit is that a single set of tables (by choice of initial state)
describes a process to parse different languages (or sub-phrases) while sharing
common specification elements.

Practical language definitions generally make use of a few extra concepts:
* Attribute-synthesis rules show how to construct a semantic value for a sentence in the language.
* Operator-precedence rules enable deterministic parsing decisions in the face of CFG ambiguities.


# Operator Precedence and Associativity (OP&A)

Classically, operator-precedence and associativity was an entire parsing method.
In more recent times, it's normally an adjunct for resolving the parsing conflicts
that come up in ambiguous context-free grammars.

The method considers two broad categories of conflict:
* Shift/Reduce: a priority decision between operators to the left and right.
* Reduce/Reduce: a choice between production rules applicable at the same point.

If explicit declarations are not provided, the grammar is
ambiguous -- at least with respect to the power of the  used.
There are ways to deal with ambiguity, but for parsers the convention
is to shift on a shift-reduce conflict (i.e. assume right-associativity) and, in
the event of a reduce/reduce conflict, to use the earliest-defined rule.


# Attribute Synthesis

The rules are annotated with instructions for how to perform a semantic transduction
of the implied parse tree. In current Booze-Tools philosophy, there are a few broad
categories of transduction steps. These include:

* Renaming: These can usually be optimized out of parse tables.
* Bracketing: Technically a special-case message parameterized only by a single stack offset.
* Message: Carries a message and a list of stack offsets for semantic parameters to the message.
* Accept-Language: Implied in augmentation-rules used for generating parse tables.


# A note on symbology:

People adopt various style-rules for representing the symbols in their grammars. These strictures add
a useful layer of redundancy. At the moment there's no support for it, but eventually it could be as
simple as adding a predicate for whether something *looks* like a non-terminal. A related

"""

import collections, itertools
from typing import Optional, NamedTuple, Hashable, Sequence, Union, Protocol
from ..support import foundation, pretty


LEFT, RIGHT, NONASSOC, BOGUS = object(), object(), object(), object()

class Fault(ValueError):
	""" Generic exception thrown by the generic handler. """

class FaultHandler(Protocol):
	"""
	This generic handler just raises exceptions.
	More sophisticated handlers might do something more sophisticated.
	This is not the final form of the protocol.

	It might be nice to get a more complete read-out of problems, but that would mean
	inventing some sort of document structure to talk about faults in grammars.
	Then again, maybe that's in the pipeline.
	"""
	def rule_produces_bogon(self, rule, symbol):
		raise Fault("Rule at %r produces bogus terminal(s) %r." % (rule.provenance, symbol))
	
	def epsilon_right_recursion(self, rule):
		raise Fault("Rule at %r produces epsilon right-recursion."%rule.provenance)
	
	def nonterminal_given_precedence(self, symbol):
		raise Fault("Nonterminal %r included in precedence declaration."%symbol)
	
	def bad_prec_sym(self, rule):
		raise Fault("Rule at %r has an explicit precedence-symbol without a defined precedence level.", rule.provenance)
	
	def ill_founded_symbols(self, symbols):
		raise Fault("Ill-Founded Symbols: %r."%symbols)
	
	def unreachable_symbols(self, symbols):
		raise Fault("Unreachable Symbols: %r."%symbols)
	
	def duplicate_rules(self, rules):
		raise Fault("Duplicated rules at %r."%', '.join(map(str, (r.provenance for r in rules))))
	
	def precedence_redeclared(self, symbol, first, extras):
		raise Fault("Precedence declared twice on symbol %r."%symbol)

	def self_recursive_loop(self, symbol):
		# FIXME: convey the location of the problem?
		raise Fault("Symbol %r may be replaced by itself in a recursive loop."%symbol)

	def mutual_recursive_loop(self, symbols):
		# FIXME: convey the locations of the problematic rules?
		raise Fault("Symbols %r may be replaced by one another in a mutually-recursive loop."%symbols)

	def epsilon_mutual_recursion(self, symbols):
		# FIXME: convey the locations of the problematic rules?
		raise Fault("Symbols %r form a mutually-recursive epsilon loop."%symbols)

class SimpleFaultHandler(FaultHandler):
	""" Protocols cannot be instantiated, so here's a simple way to get "raise for everything" behavior. """
	pass


class SemanticAction(NamedTuple):
	"""
	Designate semantics for the attribute synthesis portion of the rule specification.
	Consists of an ostensible *message* and a list of indices within the right-hand-side from
	which to gather arguments to that message. The indices are zero-indexed from the left.
	At this level, we don't interpret messages further: That's for the driver to handle.
	"""
	message: Optional[Hashable]
	indices: Sequence[int]

class Rule(NamedTuple):
	"""
	Arbitrary plain-jane BNF rule.

	lhs: The "left-hand-side" non-terminal symbol which is declared to produce...
	rhs: this "right-hand-side" sequence of symbols.
	prec_sym: Optional explicit rule precedence symbol. If missing, may be inferred.
	action:
		* If an integer, then the significant-symbol index for a renaming or bracketing rule.
		* If a SemanticAction object, then as explained there.

	Unit/renaming productions are recognized and treated specially. The parser tables
	will generally optimize such rules to a zero-cost abstraction, bypassing pointless
	extra stack activity and leaving out needless extra states. Certain very rare
	constructions may defeat the optimization by rendering it unsound. The tables will
	be constructed *correctly* in any event.

	provenance: Use this to indicate the provenance of the rule. It can be a source
	line number, or whatever else makes sense in your application.
	"""
	lhs: Hashable  # What non-terminal symbol...
	rhs: tuple[Hashable, ...]  # ...produces what string of other symbols?
	prec_sym: Optional[Hashable]  # If this rule has explicit precedence, what symbol represents its precedence level?
	action: Union[int, SemanticAction]  # What to do for attribute synthesis when recognizing this rule?
	provenance: object  # Where did this rule come from? We need to know for reporting and debugging.
	
	def __str__(self):
		return "%s -> %s"%(self.lhs, ' '.join(map(str, self.rhs)))
	
	def is_rename(self):
		return len(self.rhs) == 1 and self.action == 0
	
	def assert_valid_action(self):
		""" If this check fails, it's held to be a bug in whatever code created the rule object. """
		arity = len(self.rhs)
		if isinstance(self.action, SemanticAction):
			assert all(p<arity for p in self.action.indices), self.action
		else:
			assert isinstance(self.action, int) and 0 <= self.action < arity
	
	def as_dotted(self, position):
		rhs = [str(s) for s in self.rhs]
		rhs.insert(position, pretty.DOT)
		return str(self.lhs)+' -> '+(" ".join(rhs))

class OperatorPrecedenceSupport:
	""" Coherent bits that deal with operator-precedence parsing. This decides most shift-reduce conflicts. """
	
	def __init__(self):
		# The fundamental essentials:
		self.token_precedence = {}
		self.level_assoc = []
		
		# Now for some error-reporting help:
		self.level_provenance = []  # Used for error-reporting.
		self.bogons = set()
		self.extra_declarations = collections.defaultdict(list)

	def assoc(self, direction, symbols, provenance=None):
		assert direction in (LEFT, NONASSOC, RIGHT, BOGUS)
		assert symbols
		level = foundation.allocate(self.level_assoc, direction)
		self.level_provenance.append(provenance)
		for symbol in symbols:
			if symbol in self.token_precedence:
				self.extra_declarations[symbol].append(provenance)
			else:
				self.token_precedence[symbol] = level
	
	def infer_prec_sym(self, rhs):
		"""
		If a rule without an explicit precedence declaration is involved in a shift/reduce conflict,
		the parse table generation algorithm will call this to decide which symbol represents the
		precedence of this right-hand-side.

		As a slight refinement to the BISON approach, this returns the first terminal with an
		assigned precedence, as opposed to the first terminal symbol whatsoever. Often that's
		a distinction without a difference, but when it matters I think this makes more sense.
		"""
		for symbol in rhs:
			if symbol in self.token_precedence:
				return symbol

	def validate(self, fault_handler:FaultHandler, rules):
		for symbol, extras in self.extra_declarations.items():
			first = self.level_provenance[self.token_precedence[symbol]]
			fault_handler.precedence_redeclared(symbol, first, extras)
		for rule in rules:
			if rule.lhs in self.token_precedence:
				fault_handler.nonterminal_given_precedence(rule)
			if rule.prec_sym and rule.prec_sym not in self.token_precedence:
				fault_handler.bad_prec_sym(rule)
			for symbol in rule.rhs:
				if self.token_precedence.get(symbol,None) is BOGUS:
					fault_handler.rule_produces_bogon(rule, symbol)
	
	def determine_rule_precedence(self, rule):
		prec_sym = rule.prec_sym or self.infer_prec_sym(rule.rhs)
		if prec_sym: return self.token_precedence[prec_sym]
	
	def decide_shift_reduce(self, symbol, rule):
		try: sp = self.token_precedence[symbol]
		except KeyError: return None
		rp = self.determine_rule_precedence(rule)
		if rp is None: return None
		if rp < sp: return LEFT
		# NB: Bison and Lemon both treat later declarations as higher-precedence,
		# which is unintuitive, in that you perform higher-precedence operations
		# first so it makes sense to list them first. Please excuse my dear aunt Sally!
		if rp == sp: return self.level_assoc[rp]
		return RIGHT


class ContextFreeGrammar:
	"""
	Attributed context-free grammar with operator-precedence and associativity declarations.
	
	This object follows a builder pattern: You're expected to construct an empty grammar,
	give it a bunch of declarations, and then start calculating based on the completed grammar.
	"""
	def __init__(self):
		# Context-free bits
		self.symbols = set()
		self.rules:list[Rule] = []
		self.start = []  # The start symbol(s) may be specified asynchronously to construction or the rules.
		self.symbol_rule_ids = {}
		self.ops = OperatorPrecedenceSupport()
		# A reverse mapping from right-hand side symbols to rule IDs makes various algorithms easier/faster:
		self.mentions = collections.defaultdict(list)
	
	def display(self):
		head = ['', 'Symbol', 'Produces', 'Using']
		body = [[i, rule.lhs, rule.rhs, rule.action] for i,rule in enumerate(self.rules)]
		pretty.print_grid([head] + body)

	def add_rule(self, rule):
		"""
		This is your basic mechanism to add BNF rules.
		It's responsible for various bits of internal accounting.
		:return: Nothing.
		"""
		rule.assert_valid_action()
		self.symbols.add(rule.lhs)
		self.symbols.update(rule.rhs)
		
		if rule.lhs not in self.symbol_rule_ids:
			# Don't use a defaultdict; we can't be adding keys by checking for them.
			self.symbol_rule_ids[rule.lhs] = []
		sri = self.symbol_rule_ids[rule.lhs]
		rule_id = foundation.allocate(self.rules, rule)
		sri.append(rule_id)
		for symbol in rule.rhs:
			self.mentions[symbol].append(rule_id)

	def augmented_rules(self) -> list:
		"""
		The LR family of algorithms are normally explained in terms of an augmented grammar:
		It has a special "accept" rule which just expands to the start symbol for the ordinary
		context-free grammar at issue. There's a good reason for this: It makes the algorithms
		work properly without a lot of special edge cases to worry about. However, the "accept"
		non-terminal ought to be imaginary: it can't appear on the right, one never reduces the
		"accept" symbol, and I object to it consuming space in the GOTO table as it typically
		does in implementations that rely on a first-class rule (with left-hand side symbol).

		This object exists to clarify interaction with the "augmented" rule set.
		NB: Since we support multiple "start" symbols, there are corresponding "accept" rules.
		"""
		assert self.start
		return [rule.rhs for rule in self.rules] + [[language] for language in self.start]
	
	def initial(self) -> range:
		""" The range of augmented-rule indices corresponding to the list of start symbols. """
		first = len(self.rules)
		return range(first, first+len(self.start))
	
	def assoc(self, direction, symbols, provenance=None):
		self.ops.assoc(direction, symbols, provenance)
		
	def decide_shift_reduce(self, symbol, rule_id):
		return self.ops.decide_shift_reduce(symbol, self.rules[rule_id])
	
	def apparent_terminals(self) -> set:
		""" Of all symbols mentioned, those without production rules are apparently terminal. """
		# FIXME: This could be a pluggable strategy, and MAY be misplaced. Validation and table generation need to know.
		return self.symbols - self.symbol_rule_ids.keys()
	
	def bipartite_closure(self, root_symbols) -> set:
		"""
		The algorithm to find epsilon symbols and well-founded symbols is basically the same,
		with different initial conditions. It's a bipartite propagation, alternating between
		disjuncts (symbols) and conjuncts (rules). By factoring this out and focusing on the
		graph structure of the problem (and also getting some sleep) a decent solution arose.
		"""
		# Structuring the problem in bipartite-graph form, as explained in the docstring.
		# The transitive closure involves some stateful cleverness:
		# Rely on the foundation library for proper abstraction:
		
		def successors(symbol):
			for rule_id in self.mentions[symbol]:
				remain[rule_id] -= 1
				if remain[rule_id] == 0:
					yield self.rules[rule_id].lhs
		
		remain = [len(rule.rhs) for rule in self.rules]
		return foundation.transitive_closure(root_symbols, successors)
	
	def find_epsilon(self) -> set:
		""" Which symbols may produce the empty string? """
		return self.bipartite_closure(rule.lhs for rule in self.rules if not rule.rhs)
	
	def find_first(self):
		"""
		Particularly the LL school of grammar processing uses first-sets.
		These tables answer the question: For each nonterminal symbol,
		what terminal symbols could it possibly start with?
		
		The FIRST set of a nonterminal is the union of the FIRST sets of each right-hand side.
		Without epsilon symbols, the FIRST set of a right-hand side is that of its first symbol.
		With them, you need to include all symbols up to the first non-epsilon symbol.
		
		Anyway, this is a job for a strongly_connected_components,
		in part because it also produces a topological sort which
		allows for a quick determination of the level sets.
		"""
		
		# Structure the problem as a graph, called ``first``
		first = collections.defaultdict(set)
		epsilon = self.find_epsilon()
		for rule in self.rules:
			for symbol in rule.rhs:
				first[rule.lhs].add(symbol)
				if symbol not in epsilon: break
		
		# Gather up the answer in topological order:
		for component in foundation.strongly_connected_components_hashable(first):
			# All the members of a component share a first-set, here called ``f``.
			f = set()
			for symbol in component:  # Not sure about performance with large SCCs.
				f.update(*(first[x] for x in first[symbol]))
				first[symbol] = f
			f.difference_update(self.symbol_rule_ids) # Thus subtracting out nonterminals
		return first
	
	def assert_well_founded(self, fault_handler: FaultHandler):
		"""
		Here, "well-founded" means "can possibly produce a finite sequence of terminals."
		The opposite is called "ill-founded".

		Here are two examples of ill-founded grammars:
			S -> x S   -- This is ill-founded because there's always one more S.
			A -> B y;  B -> A x     -- Similar, but with mutual recursion.

		A terminal symbol is well-founded. So is an epsilon symbol, since zero is finite.
		A rule with only well-founded symbols in the right-hand side is well-founded.
		A non-terminal symbol with at least one well-founded rule is well-founded.
		Induction applies. That induction happens to be called ``bipartite_closure``.

		A grammar with only well-founded symbols is well-founded.
		"""
		well_founded = self.bipartite_closure(self.apparent_terminals() | self.find_epsilon())
		ill_founded = self.symbol_rule_ids.keys() - well_founded
		if ill_founded:
			fault_handler.ill_founded_symbols(ill_founded)
	
	def assert_no_orphans(self, fault_handler:FaultHandler):
		"""
		Every symbol should be reachable from the start symbol(s).
		This is a simple transitive closure.
		"""
		produces = collections.defaultdict(set)
		for rule in self.rules: produces[rule.lhs].update(rule.rhs)
		unreachable = self.symbols - foundation.transitive_closure(self.start, produces.get)
		if unreachable:  # NB: Bogons are not among self.symbols.
			fault_handler.unreachable_symbols(unreachable)
			
	def assert_no_rename_loops(self, fault_handler:FaultHandler):
		""" If a symbol may be replaced by itself (possibly indirectly) then it is diseased. """
		renames = collections.defaultdict(set)
		for rule in self.rules:
			if len(rule.rhs) == 1:
				if rule.lhs == rule.rhs[0]:
					fault_handler.self_recursive_loop(rule.lhs)
				else:
					renames[rule.lhs].add(rule.rhs[0])
		for component in foundation.strongly_connected_components_hashable(renames):
			if len(component) > 1: fault_handler.mutual_recursive_loop(component)
	
	def assert_no_epsilon_loops(self, fault_handler:FaultHandler):
		""" Epsilon Left-Self-Recursion is OK. All other recursive-epsilon-loops are pathological. """
		epsilon = self.find_epsilon()
		reaches = collections.defaultdict(set)
		for rule in self.rules:
			epsilon_prefix = list(itertools.takewhile(epsilon.__contains__, rule.rhs))
			if not epsilon_prefix: continue
			if epsilon_prefix[0] == rule.lhs: epsilon_prefix.pop(0)
			if rule.lhs in epsilon_prefix:
				# Somehow this case seems qualitatively different.
				fault_handler.epsilon_right_recursion(rule)
			reaches[rule.lhs].update(epsilon_prefix)
		for component in foundation.strongly_connected_components_hashable(reaches):
			if len(component) > 1:
				fault_handler.epsilon_mutual_recursion(component)
	
	def assert_no_duplicate_rules(self, fault_handler:FaultHandler):
		for symbol, rule_ids in self.symbol_rule_ids.items():
			inverse = collections.defaultdict(list)
			for r in rule_ids:
				inverse[tuple(self.rules[r].rhs)].append(r)
			for rs in inverse.values():
				if len(rs) > 1:
					fault_handler.duplicate_rules([self.rules[r] for r in rs])
	
	def validate(self, fault_handler=SimpleFaultHandler(), allow_duplicate_rules=False):
		"""
		Calls the fault handler with every identified fault. The default fault handler
		raises an exception (derived from Fault, above) for the first error noticed.
		"""
		self.ops.validate(fault_handler, self.rules)
		self.assert_well_founded(fault_handler)
		self.assert_no_orphans(fault_handler)
		self.assert_no_rename_loops(fault_handler)
		self.assert_no_epsilon_loops(fault_handler)
		if not allow_duplicate_rules:
			self.assert_no_duplicate_rules(fault_handler)
			




"""
Parsing Interface Definitions
"""

END_OF_TOKENS = '<END>' # An agreed artificial "end-of-text" terminal-symbol.
ERROR_SYMBOL = '$error$' # An agreed "error" symbol.
# Note that the scanner should NEVER emit either of the above two symbols.
# However, the error symbol may appear in the right-hand side of a production rule.

class ParseError(ValueError): pass
class UnexpectedTokenError(ParseError): pass
class UnexpectedEndOfTextError(ParseError): pass

class ParseErrorListener:
	"""
	Implement this interface to report/respond to parse errors.
	For the moment I'm assuming you have a handle to the scanner so you
	can get the input-file location of error events...
	"""
	def unexpected_token(self, kind, semantic, pds):
		"""
		The parser has just been given a bogus token.
		It will enter recovery mode next.
		`kind` and `semantic` are whatever the scanner provided.
		`pds` is the state of the push-down automaton at the point of error.
		"""
		raise UnexpectedTokenError(kind, semantic, pds)
	
	def unexpected_eof(self, pds):
		"""
		The parser ran out of tokens unexpectedly.
		`pds` is the state of the push-down automaton at the point of error.
		"""
		raise UnexpectedEndOfTextError(pds)
	
	def will_recover(self, tokens):
		"""
		The parser has seen a token sequence sufficient to resynchronize.
		`tokens` is that sequence. The parser will next commit to this
		recovery. (Perhaps there should be a way to prevent it?)
		The return value from this method will appear as the semantic content
		of the "error" position in the error rule that was ultimately chosen.
		"""
	
	def did_not_recover(self):
		"""
		The parser ran out of tokens while in error-recovery mode, and was
		unable to recover.
		"""
	
	def cannot_recover(self):
		"""
		The parser attempted to enter recovery mode, but there are no
		recoverable states on the parse stack, so recovery is impossible.
		
		Default behavior is `return self.did_not_recover()`
		"""
		return self.did_not_recover()
	
	def exception_parsing(self, ex:Exception, message, args):
		"""
		Q: If a combining function raises an exception, what should happen?
		A: It depends.
		
		Maybe the exception should not happen: some extra context might help
		you reproduce and debug the problem. Log the context and re-raise.
		
		Maybe certain exceptions represent non-fatal conditions, but you'd
		rather separate policy from mechanism. Deal with it and return the
		semantic value that should replace the aborted attribute-synthesis.
		"""
		raise ex from None # Hide the catch-and-rethrow from the traceback.

class HandleFindingAutomaton:
	"""
	This interface captures the operations needed to perform table-driven parsing, as well as a modicum
	of reasonable error reporting. Again, no particular structure or organization is implied.
	"""
	
	# The first several methods are involved in actually finding handles.
	def get_initial(self, language) -> int: raise NotImplementedError(type(self), 'return the initial state id for the selected language, which by the way is usually `None `.')
	def get_translation(self, symbol) -> int: raise NotImplementedError(type(self, 'Because scanners should not care the order of terminals in the parse table. Zero is reserved for end-of-text.'))
	def get_action(self, state_id:int, terminal_id) -> int: raise NotImplementedError(type(self), 'Positive -> successor state id. Negative -> rule id for reduction. Zero -> error.')
	def get_goto(self, state_id:int, nonterminal_id) -> int: raise NotImplementedError(type(self, 'return a successor state id.'))
	def get_breadcrumb(self, state_id:int) -> str: raise NotImplementedError(type(self), 'This is used in error reporting. Return the name of the symbol that shifts into this state.')
	def interactive_step(self, state_id:int) -> int: raise NotImplementedError(type(self), 'Return the reduce instruction for interactive-reducing states; zero otherwise.')

	# The next few are involved in reducing handles. (Yes, this interface is therefore overgrown.)
	def get_rule(self, rule_id:int) -> tuple: raise NotImplementedError(type(self), 'return a (nonterminal_id, length, constructor_id, view) quad.')
	def get_constructor(self, constructor_id) -> object: raise NotImplementedError(type(self), 'return whatever will make sense to the corresponding combiner.')
	def each_constructor(self) : raise NotImplementedError(type(self), "Involved in binding parser tables to drivers. Yield pairs of <constructor, set of mentions>.")
	
	# These next two methods are in support of GLR parsing:
	def get_split_offset(self) -> int: raise NotImplementedError(type(self), "Action entries >= this number mean to split the parser.")
	def get_split(self, split_id:int) -> list: raise NotImplementedError(type(self), "A list of parse actions of the usual (deterministic) form.")


class AbstractParser:
	"""
	Before I get too deep into it, let's lay out the general structure of a generalized parse:
	"""
	def __init__(self, table: HandleFindingAutomaton, combine, language=None):
		""" Please note this takes a driver not a combiner: it does its own selection of arguments from the stack. """
		self._table = table
		self._combine = combine
		self._nr_states = table.get_split_offset()
		self.reset(table.get_initial(language))
	
	def reset(self, initial_state):
		""" Configure the initial stack situation for the given initial automaton state. """
		raise NotImplementedError(type(self))
	
	def consume(self, terminal, semantic):
		""" Call this from your scanning loop. """
		raise NotImplementedError(type(self))

	def finish(self) -> list:
		"""
		Call this after the last token to wrap up and
		:return: a valid semantic value for the parse.
		"""
		raise NotImplementedError(type(self))



"""
Algorithms and data structures for constructing, determinizing, and minimizing
finite state machines. These make no attempt at clever internal storage: they
follow the maxim "When in doubt, use brute force."
"""

import operator
from typing import NamedTuple
from ..support import foundation, pretty
from ..scanning import charclass, charset
from .interface import Classifier, FiniteAutomaton, State, CodePoint, RuleId

class DFA(FiniteAutomaton):

	def __init__(self, *, alphabet: Classifier, initial:dict, final:dict, states:list):
		self.alphabet = alphabet
		self.width = self.alphabet.cardinality()
		self.initial = initial
		self.final = final
		metric = [len(row) for row in states]
		assert all(m==self.width for m in metric), (self.width, metric)
		self.states = states
		
	def jam_state(self): return -1
	def condition(self, condition_name) -> tuple: return self.initial[condition_name]
	def transition(self, state: State, codepoint: CodePoint) -> State: return self.states[state][self.alphabet.classify(codepoint)]
	def accept(self, state: State) -> RuleId: return self.final.get(state)
	
	def append_state(self, row) -> int:
		assert len(row) == self.width, [len(row), self.width]
		return foundation.allocate(self.states, row)
	
	def display(self):
		print('Finite Automaton:')
		self.alphabet.display()
		print('Initial:', self.initial)
		head = ['*', '', ]+list(range(self.alphabet.cardinality()))
		body = [[self.final.get(i,''), i, *[s if s >= 0 else '' for s in row]] for i, row in enumerate(self.states)]
		pretty.print_grid([head] + body)
	
	def minimize_states(self) -> "DFA":
		""" Let's try Moore's Algorithm: It's easier to get right. """
		buckets = []
		partition_id_of_state = []
		def establish_initial_partitioning():
			""" Every state goes into a partition numbered according to its finality: """
			finality = {}
			for n in range(len(self.states)):
				rule_id = self.final.get(n)
				if rule_id not in finality: finality[rule_id] = foundation.allocate(buckets, [])
				b = finality[rule_id]
				partition_id_of_state.append(b)
				buckets[b].append(n)
			pass
		def translate(q) -> tuple: return tuple(-1 if x < 0 else partition_id_of_state[x] for x in self.states[q])
		def refine_partitions() -> bool:
			splitting_happened = False
			for b, bucket in enumerate(buckets):
				if len(bucket) < 2: continue
				exemplar = translate(bucket[0])
				same, different = [], []
				for q in bucket: (same if all(map(operator.eq, exemplar, translate(q))) else different).append(q)
				if different:
					buckets[b] = same
					n = foundation.allocate(buckets, different)
					for q in different: partition_id_of_state[q] = n
					splitting_happened = True
			return splitting_happened
		def interpretation() -> DFA:
			return DFA(
				alphabet=self.alphabet,
				initial={condition: (partition_id_of_state[q0], partition_id_of_state[q1]) for condition, (q0, q1) in self.initial.items()},
				final = {partition_id_of_state[q]:rule_id for q,rule_id in self.final.items()},
				states = [translate(bucket[0]) for bucket in buckets],
			)
		establish_initial_partitioning()
		while refine_partitions(): pass
		return interpretation()
		
	def minimize_alphabet(self) -> "DFA":
		assert isinstance(self.alphabet, charclass.SimpleClassifier)
		ec = foundation.EquivalenceClassifier()
		classes = [ec.classify(column) for column in zip(*self.states)]
		return DFA(
			alphabet=charclass.MetaClassifier(self.alphabet.bounds, classes),
			initial=self.initial,
			final=self.final,
			states=list(zip(*ec.exemplars)),
		)
	
	def stats(self):
		Q = len(self.states)
		W = self.width
		X = sum(sum(x != -1 for x in state) for state in self.states)
		print('DFA has %d states and %d character classes, using %d cells.'%(Q, W, Q*W))
		print('%d non-error cells, or %0.2f%%'%(X, 100*X/(Q*W)))
	
	def make_csv(self, pathstem):
		pretty.write_csv_grid(pathstem + '.dfa.csv', [
			[q, self.final.get(q), '', *row]
			for q, row in enumerate(self.states)
		])
		pass

class NFA:
	def __init__(self):
		self.states, self.initial, self.final = [], {}, {}
		self.all_bounds = {0} # There is at least an "end-of-file" marker, which is distinct from all else.

	class Edge(NamedTuple):
		label: list
		target: int

	class Node(NamedTuple):
		edges: list
		epsilons: set
		rank: int
	
	def new_node(self, rank) -> int:
		return foundation.allocate(self.states, NFA.Node(edges=[], epsilons=set(), rank=rank))
	def condition(self, name):
		if name not in self.initial: self.initial[name] = (self.new_node(0), self.new_node(0))
		return self.initial[name]
	def link(self, src :int, dst :int, label :list):
		self.states[src].edges.append(NFA.Edge(label, dst))
		self.all_bounds.update(label)
	def link_epsilon(self, src :int, dst :int): self.states[src].epsilons.add(dst)
	def link_condition(self, main_condition, included_condition):
		"""
		Precondition: Both conditions exist, and are known to the NFA `self`.
		Postcondition: It is as if all the rules in `included_condition` are also defined in `main_condition`,
		and in the same relative order.
		"""
		for src, dst in zip(self.initial[main_condition], self.initial[included_condition]):
			self.link_epsilon(src, dst)
	def subset_construction(self) -> DFA:
		"""
		The standard plan to convert an NFA to a DFA, in plain English, is to consider
		that a deterministic state represents a particular and distinct subset of NFA states.
		That general plan is embodied by the interplay between the generic BreadthFirstTraversal(...)
		and `visit(key)`: the "key" is composed principally of a `frozenset` of NFA state ID numbers.
		
		This module also supports "rule ranks" -- about which more is written elsewhere. To implement
		the concept, an extra bit of data rides along with the subset: specifically, a rank number.
		"""
		def close(ns, min_rank:int):
			assert all(isinstance(n, int) for n in ns)
			tc = foundation.transitive_closure(ns, lambda n :self.states[n].epsilons)
			ns = frozenset(n for n in tc if self.states[n].rank >= min_rank)
			return ns, min((self.states[n].rank for n in ns), default=0)
		def visit(key):
			ns, min_rank = key
			active, targets = [], []
			final = ns & self.final.keys()
			if final:
				min_rank = max(self.states[n].rank for n in final)
				dfa.final[len(dfa.states)] = min(self.final[n] for n in final if self.states[n].rank == min_rank)
			for n in ns:
				node = self.states[n]
				if node.rank >= min_rank:
					for e in node.edges:
						active.append(list(charset.expand(e.label, all_bounds)))
						targets.append(e.target)
			# Now, if active[i][j], then targets[i] participates in class[j].
			delta, prior = [], None
			for column in zip(*active):
				register = [t for r,t in zip(column, targets) if r]
				if register != prior: # Save a lot of transitive closure operations...
					prior = register
					subset, subrank = close(register, min_rank)
					successor = bft.lookup((subset, subrank)) if subset else -1
				delta.append(successor)
			dfa.append_state(delta or [-1]*dfa.width)
		# Main routine:
		all_bounds = sorted(self.all_bounds|{-1})
		dfa = DFA(alphabet=charclass.SimpleClassifier(all_bounds[1:]), initial={}, final={}, states=[])
		bft = foundation.BreadthFirstTraversal()
		def initial(q:int): return close([q], 0)
		dfa.initial = {k :(bft.lookup(initial(a)), bft.lookup(initial(b))) for k, (a, b) in self.initial.items()}
		bft.execute(visit)
		return dfa



"""
Trees, version two.

The current implemented notion is that the tree contains only the tree; semantic data is separate.
Practical tree-passes descend from `class TreePass` and expose attributes to describe semantic values.
This is still far from the final form, but it has enabled some interesting discoveries.
"""

import abc

class RankedAlphabet:
	"""
	A ranked alphabet is a couple (F,Arity) where F is a finite set and Arity is a mapping from F into N.
		-- TATA, page 15.
	
	This version compels symbols to be categorized (many-to-many) to the extent they'll be used for anything.
	I don't yet bother with a notion of meta-variables.
	"""
	
	def __init__(self, *categories:str):
		self.__categories = {c:set() for c in categories}
		assert all(c.isidentifier() for c in categories)
		self.__symbols = {}
	
	def __getitem__(self, item):
		return self.__symbols[item]
		
	def categorize(self, category_name, *symbol_names):
		""" Declares membership of symbols in categories. Contributes to the run-time validation layer. """
		self.__categories[category_name].update(self.__symbols[s] for s in symbol_names)
		
	def symbol(self, name:str, *categories:str, **slots) -> type:
		"""
		This function creates/registers a new type of symbol as a Python class derived from BaseTerm (below).
		Keyword-arguments are field-name = category-name, except that category-names can have a multiplicity suffix.
		The suffixes are just the ordinary ? * + characters understood from regular expressions.
		Constructing an instance will assert that it's been constructed from the correct field types.
		Optional fields can be None. Multiple fields should be a list or tuple.
		"""
		
		def make_check(cat:str):
			if cat.endswith('?'): return _check_optional(self.__categories[cat[:-1]])
			elif cat.endswith("*"): return _check_zero_or_more(self.__categories[cat[:-1]])
			elif cat.endswith("+"): return _check_one_or_more(self.__categories[cat[:-1]])
			else: return _check_one(self.__categories[cat])
		
		assert name not in self.__symbols
		assert name.isidentifier()  # Keep this; we dispatch to methods by these names.
		it = self.__symbols[name] = type(name, (BaseTerm,), {
			"__slots__": tuple(slots.keys()),
			"_arity_": len(slots),
			"_type_predicates_": tuple(make_check(cat) for cat in slots.values()),
			"_appearance_": '<%s/%d>'%(name,len(slots)),
			"_alphabet_":self,
		})
		for c in categories:
			self.__categories[c].add(it)
		return it
	
def _check_one(category): return lambda x: type(x) in category
def _check_one_or_more(category): return lambda xs: xs and all(type(x) in category for x in xs)
def _check_zero_or_more(category): return lambda xs: all(type(x) in category for x in xs)
def _check_optional(category): return lambda x: x is None or type(x) in category


class BaseTerm:
	"""
	Thus, a term t ∈ T(F,X) may be viewed as a finite ordered ranked tree,
	the leaves of which are labeled with variables or constant symbols and
	the internal nodes are labeled with symbols of positive arity, with
	out-degree equal to the arity of the label...
		-- TATA, page 15.
	
	We confuse terms and trees...
		-- TATA, page 16.

	This is the base-class for generated symbol types,
	with definitions for interesting magic-methods.
	It's somewhere between a dataclass and a namedtuple.
	"""
	_arity_: int
	__slots__: tuple
	_type_predicates_: tuple
	_appearance_: str
	def __setattr__(self, key, value): raise TypeError
	def __delattr__(self, item): raise TypeError
	def __init__(self, *args):
		if __debug__ and len(args) != self._arity_:
			raise TypeError(self._appearance_+" got %d arguments"%len(args))
		for f,v,check in zip(self.__slots__, args, self._type_predicates_):
			if __debug__ and not check(v):
				raise TypeError("%s: Field %r failed well-formed-ness check with value %r"%(self._appearance_,f,v))
			object.__setattr__(self, f,v)
	def __str__(self): return self._appearance_
	def __iter__(self):
		return (getattr(self, s) for s in self.__slots__)


class TreePass(abc.ABC):
	"""
	Compiler passes are callable:
	They implement a simple form of double-dispatch by calling a method named
	for the type of the first argument, and with the term again as first argument.
	All remaining arguments are passed through unexamined.
	This provides an easy way to write micro-passes,
	but it means the per-symbol methods must have explicit recursive calls
	if you intend for processing to continue.
	
	(Incidentally, see also the function "post_order".)
	"""
	@abc.abstractmethod
	def _unhandled_(self, term, *args, **kwargs):
		""" Deal with unknown symbols here. """
		raise NotImplementedError(type(self))
	
	def __call__(self, term, *args, **kwargs):
		method = getattr(self, term.__class__.__name__, self._unhandled_)
		return method(term, *args, **kwargs)
	
	pass

class SparseWalk(TreePass):
	""" Calls your methods on interesting nodes; does not build a new tree. """
	def _unhandled_(self, term, *args, **kwargs):
		for child in term:
			self(child, *args, **kwargs)

class SparseRewrite(TreePass):
	def _unhandled_(self, term, *args, **kwargs):
		"""
		This is the generic structural-copy method.
		It conveniently also works for lists and tuples.
		It also goes out of its way to share identical sub-trees.
		Smart pruning would be a nice feature here.
		The very smartest pruning would take advantage of a bottom-up DFA generated from the translation plan,
		marking nodes upon creation with their eligibility for subsequent traversals. But that's a long way off.
		Also, this highlights the need for traceability.
		"""
		kids = tuple(term)
		xlat = tuple(self(child, *args, **kwargs) for child in kids)
		if all(a is b for a,b in zip(kids, xlat)):
			return term
		else:
			return type(term)(*xlat)

class StrictPass(TreePass):
	def _unhandled_(self, node, *args, **kwargs):
		""" Strict passes must implement something for every symbol. """
		appearance = str(node) if isinstance(node, BaseTerm) else str(type(node))
		raise RuntimeError("class %s neglects to handle symbol %s"%(type(self), appearance))


def post_order(visitor, *, otherwise=lambda x:x):
	"""
	Logic for a simple bottom-up left-to-right micro-pass:

	The visitor is assumed to have methods named for symbols.
	These are called in post-order with bottom-up results.

	If the algorithm encounters a term for which no method exists,
	it returns `otherwise(` that term `)` instead of delving further.
	The default `otherwise` is to return the node as-is.
	(Use otherwise=None to raise AttributeError instead.)
	Lists and tuples get a structural mapping.
	All other data types are returned as-is.
	"""
	def visit(term):
		if isinstance(term, BaseTerm):
			try: method = getattr(visitor, term.__class__.__name__)
			except AttributeError:
				if otherwise is None: raise
				else: return otherwise(term)
		elif isinstance(term, (list, tuple)):
			method = type(term)
		else:
			return term
		return method(*(visit(t) for t in term))
	return visit



"""
Provide a mechanism to read the compacted scanning and parsing tables,
prepared using the algorithms in .compaction.py. If you were to build
a runtime system for another host language, you'd have to port these.
"""

from typing import Callable, Iterable
from .interface import ScanAction
from boozetools.scanning.interface import FiniteAutomaton
from boozetools.scanning import charclass
from boozetools.parsing.interface import HandleFindingAutomaton

def displacement_function(otherwise:callable, *, offset, check, value) -> callable:
	"""
	(make a) Reader for a perfect-hash such as built at compaction.encode_displacement_function.
	If the queried cell is absent, the `otherwise` functional-argument provides the back-up plan.
	I'll admit that form is a bit unusual in Python, but it DOES encapsulate the interpretation.
	"""
	size = len(check)
	def fn(row_id:int, column:int) -> int:
		index = offset[row_id] + column
		if 0 <= index < size and check[index] == row_id:
			return value[index]
		else:
			return otherwise(row_id, column)
	return fn

def boolean_field(row_class, col_class, offset, check) -> Callable[[int, int], bool]:
	"""
	Make a reader for a boolean field. It's used in a couple places.
	They're compressed by equivalence-classification along both dimensions.
	"""
	size = len(check)
	def fn(row:int, col:int) -> bool:
		try: rc, cc = row_class[row], col_class[col]
		except IndexError: return False # The bogus-token is out of range.
		index = offset[rc] + cc
		return 0 <= index < size and check[index] == rc
	return fn

def scanner_delta_function(*, exceptions:dict, bg:dict) -> callable:
	"""
	This function came out of some observations about the statistics of scanner delta functions.
	The best way to understand this function is by reference to the following URL:
	https://github.com/kjosib/booze-tools/issues/8
	
	The basic idea is to first check a table of "unusual" entries in the scanner's transition matrix.
	If that comes up empty, then the "background" table is essentially a bitmap combined with a pair
	of most-common entries per given row. The bitmap compresses very well by means of row- and
	column-equivalence classes; only the 1's (being the less common) need be actually represented.
	
	The very most common entry in a row is usually the error transition (-1), so instead of a
	complete list per state, only the exceptions to this rule are listed.
	"""
	zeros = dict(zip(*bg['zero']))
	broad_pattern = boolean_field(bg['row_class'], bg['col_class'], bg['offset'], bg['check'])
	def general_population(state_id:int, symbol_id:int) -> int:
		if broad_pattern(state_id, symbol_id): return bg['one'][state_id]
		else: return zeros.get(state_id, -1)
	return displacement_function(general_population, **exceptions)

def parser_action_function(*, reduce, fallback, edits) -> callable:
	background = parser_reduce_function(**reduce)
	offset, check, value = edits['offset'], edits['check'], edits['value']
	size = len(check)
	def chase(row, col):
		probe = row
		while probe >= 0:
			index = offset[probe] + col
			if 0 <= index < size and check[index] == probe: return value[index]
			probe = fallback[probe]
		return background(row, col)
	absent = len(edits['check'])
	interactive = [r if o == absent and f==-1 else 0 for r,o,f in zip(reduce['d_reduce'], edits['offset'], fallback)]
	return chase, interactive.__getitem__

def parser_goto_function(*, row_index, col_index, quotient, mark ) -> callable:
	def probe(state_id:int, nonterminal_id:int):
		r, c = row_index[state_id], col_index[nonterminal_id]
		dominant = min(r, c)
		return quotient[dominant] if dominant < mark else quotient[r+c-mark]
	return probe

def parser_reduce_function(*, d_reduce, row_class, col_class, offset, check):
	predicate = boolean_field(row_class, col_class, offset, check)
	def fn(row, col):
		dr = d_reduce[row] # Consulting the predicate costs a few cycles.
		return dr if dr and predicate(row, col) else 0 # Do it only at need.
	return fn

def scan_actions(action:dict) -> Iterable[ScanAction]:
	args = [action[what] for what in ScanAction._fields]
	return map(ScanAction, *args)

class CompactDFA(FiniteAutomaton):
	"""
	This implements the FiniteAutomaton interface (for use with the generic scanner algorithm)
	by reference to a set of scanner tables that have been built using the MacroParse machinery.
	It's not the whole story: Something else needs to interpret the rules (given here only by number),
	and rules can possibly alter the state of the scanner also.
	"""
	def __init__(self, *, dfa:dict, alphabet:dict):
		self.classifier = charclass.MetaClassifier(**alphabet)
		self.delta = scanner_delta_function(**dfa['delta'])
		self.initial = dfa['initial']
		self.final = dict(zip(dfa['final'], dfa['rule']))
	
	def jam_state(self): return -1
	def condition(self, condition_name) -> tuple:
		try: return self.initial[condition_name]
		except KeyError: raise KeyError(condition_name, set(self.initial.keys()))
	def accept(self, state_id: int) -> int: return self.final.get(state_id)

	def transition(self, current_state: int, codepoint: int) -> int:
		return self.delta(current_state, self.classifier.classify(codepoint))
	
class CompactHFA(HandleFindingAutomaton):
	"""
	This implements the HandleFindingAutomaton interface (for use with the generic parse algorithm)
	by reference to a set of parser tables that have been built using the MacroParse machinery.
	It's not the whole story: something needs to provide reduction bindings.
	"""
	def __init__(self, parse_table:dict):
		# Parts devoted to finding handles per-se
		self.get_action, self.interactive_step = parser_action_function(**parse_table['action'])
		self.get_goto = parser_goto_function(**parse_table['goto'])
		self.terminals = parse_table['terminals']
		self.__translation = {symbol:i for i,symbol in enumerate(self.terminals)}
		self.nonterminals = parse_table['nonterminals']
		self.initial = parse_table['initial']
		self.breadcrumbs = parse_table['breadcrumbs']
		
		# Parts devoted to reducing according to rules
		rule = parse_table['rule']
		self.__constructor = rule['constructor']
		self.__line_number = rule['line_number']
		self.__rule = rule['rules']
		self.get_rule = self.__rule.__getitem__
		
		# Parts devoted to generalized LR parsing
		if 'splits' in parse_table:
			self.get_split_offset = parse_table['action']['reduce']['d_reduce'].__len__ # Gets the number of states.
			self.get_split = parse_table['splits'].__getitem__
		
	def get_translation(self, symbol) -> int:
		try: return self.__translation[symbol]
		except KeyError: return len(self.terminals) # Guaranteed to trigger error-processing.
	
	def get_action(self, state_id: int, terminal_id) -> int: assert False, 'See the constructor.'
	def get_goto(self, state_id: int, nonterminal_id) -> int: assert False, 'See the constructor.'
	def get_initial(self, language) -> int: return 0 if language is None else self.initial[language]
	def get_breadcrumb(self, state_id: int) -> str:
		bcid = self.breadcrumbs[state_id]
		if bcid is None: return ''
		if bcid < len(self.terminals): return self.terminals[bcid]
		else: return self.nonterminals[bcid-len(self.terminals)]
	def interactive_step(self, state_id: int) -> int: assert False, 'See the constructor.'
	def get_split_offset(self) -> int: assert False, 'This method gets replaced if there are splits.'
	def get_split(self, split_id: int) -> list: assert False, 'See the constructor.'
	def get_rule(self, rule_id: int) -> tuple: return self.__rule[rule_id]
	def get_constructor(self, constructor_id) -> object: return self.__constructor[constructor_id]
	
	def each_constructor(self):
		mentions = [set() for _ in self.__constructor]
		for line_number, (ntid, size, cid, places) in zip(self.__line_number, self.__rule):
			if cid >= 0:
				mentions[cid].add(line_number)
		return zip(self.__constructor, mentions)


		

def _expand_rules(raw):
	"""
	Implement a proposed interpretation of a tight rules table.
	There's a related idea that might work in a C-like language where the rule IDs
	in the ACTION table are pre-translated to the offset (from the end) of their
	position in the rule data table. Then you just need a sentinel 0 at the end.
	"""
	result = []
	i = 0
	while i < len(raw):
		nonterminal_id, rhs_length, constructor_id = raw[i:i+3]
		view = []
		i += 3
		while i < len(raw) and raw[i] < 0:
			view.append(raw[i])
			i += 1
		result.append((nonterminal_id, rhs_length, constructor_id, tuple(view)))
	return result



""" Hook regular expression patterns up to method calls on a scanner object. """

from . import finite, charset
from .interface import Bindings
from .engine import Scanner, IterableScanner
from .regular import Encoder, RulePattern, analyze_pattern, let_subexpression


class Definition(Bindings):
	def __init__(self, name="MiniScan Definition", *, minimize=True):
		self.__actions = []
		self.__trails = []
		self.__subex = charset.mode_normal.new_child(name)
		self.__leaves = {}
		self.__dfa = None
		self.__nfa = finite.NFA()
		self.__minimize = minimize
		self.__awaiting_action = False
	
	def get_trailing_context(self, rule_id: int): return self.__trails[rule_id]
	
	def get_dfa(self):
		if self.__dfa is None:
			if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the final pattern!')
			self.__dfa = self.__nfa.subset_construction()
			if self.__minimize: self.__dfa = self.__dfa.minimize_states().minimize_alphabet()
		return self.__dfa
	
	def scan(self, text, *, start=None):
		return IterableScanner(text, self.get_dfa(), self, start=start)
		
	def __install_rule(self, *, action:callable, program: RulePattern, condition: (str, list, tuple)=None, rank:int=0) -> int:
		rule_id = len(self.__actions)
		self.__actions.append(action)
		self.__trails.append(program.trail_code)
		dst = self.__nfa.new_node(rank)
		self.__nfa.final[dst] = rule_id
		if condition is None or isinstance(condition, str): condition = [condition]
		else: assert isinstance(condition, (tuple, list)), type(condition)
		
		src = Encoder(self.__nfa, rank=rank, annotation=program.annotation)(program.tree, dst)
		for C in condition:
			for q, b in zip(self.__nfa.condition(C), program.bol):
				if b: self.__nfa.link_epsilon(q,src)
		return rule_id
	
	def let(self, name:str, pattern:str):
		
		let_subexpression(self.__subex, name, pattern)

	def token(self, kind:str, pattern:str, *, rank=0, condition=None):
		""" This says every member of the pattern has token kind=kind and semantic=matched text. """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:IterableScanner): yy.token(kind, yy.match())
	
	def token_map(self, kind:str, pattern:str, fn:callable, *, rank=0, condition=None):
		""" Every member of the pattern has token kind=kind and semantic=fn(matched text). """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:IterableScanner): yy.token(kind, fn(yy.match()))

	def ignore(self, pattern:str, *, rank=0, condition=None):
		""" Tell Scanner to ignore what matches the pattern. """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:IterableScanner): pass

	def on(self, pattern:str, *, condition=None, rank=0):
		"""
		For instance:
		scanner_definition.on('r#[^\n]+')(None) # Ignore comments
		@scanner_definition.on(r'[A-Za-z_]+')
		def word(yy): yy.token('word', scanner.match())
		"""
		if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the previous pattern!')
		self.__awaiting_action = True
		program = analyze_pattern(pattern, self.__subex)
		def decorator(fn):
			assert self.__awaiting_action
			self.__awaiting_action = False
			assert callable(fn)
			self.__install_rule(action=fn, program=program, condition=condition, rank=rank)
			return fn
		return decorator
	def condition(self, *condition): return ConditionContext(self, *condition)
	
	def on_match(self, yy: Scanner, rule_id:int):
		yy.less(self.get_trailing_context(rule_id))
		action = self.__actions[rule_id]
		if callable(action): action(yy)
		else: assert action is None
	
class ConditionContext:
	""" I'd like to be able to use Python's context manager protocol to simplify writing definitions of scan conditions. """
	def __init__(self, definition:Definition, *condition):
		self.__definition = definition
		self.__condition = condition
	def __enter__(self): return self
	def __exit__(self, exc_type, exc_val, exc_tb): pass
	def on(self, pattern, *, rank=0): return self.__definition.on(pattern, rank=rank, condition=self.__condition)
	def token(self, kind:str, pattern:str, *, rank=0): self.__definition.token(kind, pattern, rank=rank, condition=self.__condition)
	def token_map(self, kind:str, pattern:str, fn:callable, *, rank=0): self.__definition.token_map(kind, pattern, fn, rank=rank, condition=self.__condition)
	def ignore(self, pattern:str, *, rank=0): self.__definition.ignore(pattern, rank=rank, condition=self.__condition)

